; This file shows that the implementation of fluid variables is inconsistent
; with the implementation of shift/reset. Both implementations are part
; of Scheme48, version 1.3.
;
; To run the example, start scheme48 and do the following:
;
;   ,open fluids
;   ,open signals escapes
;   ,load =scheme48/misc/shift-reset.scm
;   ,load dynvar-scheme48-problem.scm
;
; The first line loads the structure `fluids'. The next two lines
; make available the control operators shift and reset.
; The last line loads the present code.
;
; $Id: dynvar-scheme48-problem.scm,v 1.3 2006/04/08 08:17:36 oleg Exp $

(define my-fluid (make-fluid #f))

; de-referencing of my-fluid
(define (dref) (fluid my-fluid))

; dynamic binding: execute thunk in the dynamic env obtained by binding
; my-fluid to new-val
(define (dlet new-val thunk) (let-fluid my-fluid new-val thunk))

; The first test
(define (test1)
  (dlet 1
    (lambda ()
      (list (dref) (dref)
      (dlet 2 
	(lambda ()
	  (list (dref) (dref))))))))

(display (test1)) (newline)
; (1 1 (2 2))


; The counter-example itself

; A sample function to print the current dynamic environment, twice
; (just for the heck of it).
(define (tf title)
  (display title) (newline)
  (display (dref)) (newline)
  (display (dref)) (newline))

; The first test: we capture a delimited continuation that 
; DOES include `dlet 2'
(define (test2)
  (newline) (display "test2") (newline)
  (dlet 1
    (lambda ()
      (let ((f 
	      (reset
		(dlet 2 
		  (lambda ()
		    ((shift f (begin (tf "inside shift") f)) "inside again")
		    )))))
	(tf "outside")
	(f tf)
	(tf "outside again")
))))
(test2)

; the output of the above test:
;
; test2
; inside shift
; 2
; 2
; outside
; 1
; 1
; inside again
; 2
; 2
; outside again
; 1
; 1


; The second test: we now capture a delimited continuation that does NOT
; include `dlet 2'
(define (test3)
  (newline) (display "test3") (newline)
  (dlet 1
    (lambda ()
      (let ((f 
		(dlet 2 
		  (lambda ()
	      (reset
		    ((shift f (begin (tf "inside shift") f)) "inside again")
		    )))))
	(tf "outside")
	(f tf)
	(tf "outside again")
))))
(test3)

; We observe, test2 and test3 print the same results. 
; But that is quite unexpected and at odds with the combined direct
; semantics of dynamic variables and delimited continuations.
; In test3, the dynamic environment in the case "inside again"
; (that is, when the captured continuation is reinstated)
; must be 1 rather than 2 because `dlet 2' is NOT included within the
; captured continuation.


(newline) (display "Tests from the paper, Section 4") (newline)

(display "Eq4, Section 4.1: ")
(display
  (let ((p (make-fluid 'undefined)))
    (let-fluid p 1 (lambda () (reset (fluid p))))))
(newline)
; ==> 1

(display "Eq5, Section 4.1: ")
(display
  (let ((p (make-fluid 'undefined)))
    (let-fluid p 1
      (lambda ()
	(reset
	  (let-fluid p 2
	    (lambda ()
	      (shift f (fluid p)))))))))
(newline)
; ==> 2

(display "Eq6, Section 4.2: ")
(display
  (let ((p (make-fluid 'undefined))
	(q (make-fluid 'undefined)))
    ((lambda (f)
       (let-fluid p 2
	 (lambda ()
	   (let-fluid q 20
	     (lambda ()
	       (f 0))))))
     (let-fluid p 1
       (lambda ()
	 (reset
	   (let-fluid q 10
	     (lambda ()
	       ((lambda (x) (+ (fluid p) (fluid q)))
		 (shift f f))))))))))
(newline)
; ==> 11
